/*
 * arduinolib.cpp
 *
 *  Created on: 04.11.2011
 *      Author: Johannes Liebrecht
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include<math.h>
#include <sys/time.h>
#include "arduinolib.h"

#define UDP_RX_PACKET_MAX_SIZE 24

#define FOCUS_MIN	173
#define FOCUS_MAX	724
#define ZOOM_MIN	173
#define ZOOM_MAX	724
//objective range 12 -120 mm
#define OBJECTIVE_MIN	12
#define OBJECTIVE_MAX	120

int sock, n, sockCam;
unsigned int length;
struct sockaddr_in server, from, serverCam, serverFrom;
char commandString[10];
char receiveBuffer[UDP_RX_PACKET_MAX_SIZE];
int analogInputValue[1];

void error(const char *);

/*
 * send command to arduino
 */

void sendCommand(const char command[]) {

        n = sendto(sock, command, strlen(command), 0,
                        (const struct sockaddr *) &server, length);
        if (n < 0)
                error("Sendto");
}

/*
 * wait for a packet from arduino
 * return : received packet
 */

int receivePacket() {

        n = recvfrom(sock, receiveBuffer, UDP_RX_PACKET_MAX_SIZE, 0,
                        (struct sockaddr *) &from, &length);
        if (n < 0)
                error("recvfrom");

        return atoi(receiveBuffer);
}

/*
 * get the current zoom and focus values from arduino
 */
int* getAnalogValues() {
        sendCommand("0");
        analogInputValue[0] = receivePacket();
        analogInputValue[1] = receivePacket();
        return analogInputValue;
}

/*
 * creates UDP socket
 * first : arduino ip
 * second : arduino port
 */
void createConnectionToArduino(const char ip[], int port) {
        sock = socket(AF_INET, SOCK_DGRAM, 0);
        if (sock < 0)
                error("socket");

        server.sin_family = AF_INET;
        server.sin_addr.s_addr = inet_addr(ip);
        server.sin_port = htons(port);
        length = sizeof(struct sockaddr_in);
}

/*
 * close UDP socket
 */
void disconnectConnectionToArduino() {
        close(sock);
}

/*
 * set objective zoom and wait until set
 * desiredValue : 12-120
 */
void setZoomToMm(int desiredValue) {
        assert(desiredValue>=12);
        assert(desiredValue<=120);
        sendCommand("1");
        sprintf(commandString, "%d", desiredValue);
        sendCommand(commandString);
        receivePacket();
}

/*
 * set objective focus and wait until set
 * desiredValue : 12-120
 */
void setFocusToMm(int desiredValue) {
        assert(desiredValue>=12);
        assert(desiredValue<=120);
        sendCommand("2");
        sprintf(commandString, "%d", desiredValue);
        sendCommand(commandString);
        receivePacket();
}

/*
 * set objective focus
 * !!Attention!! did not wait until objective focus is set
 * desiredValue : 12-120
 */
void setFocusToMmNoWait(int desiredValue) {
        assert(desiredValue>=12);
        assert(desiredValue<=120);
        sendCommand("5");
        sprintf(commandString, "%d", desiredValue);
        sendCommand(commandString);
}

void error(const char *msg) {
        perror(msg);
        exit(0);
}

void autoFocus() {
        int focusLow, focusHigh;
        focusLow = 12;
        focusHigh = 120;
        long max = 0;
        long temp = 0;
        int maxi = 0;

        getAnalogValues();
        if (analogInputValue[0] > ((int) (FOCUS_MAX - FOCUS_MIN) / 2) + FOCUS_MIN) {

                setFocusToMm(focusHigh);
                setFocusToMmNoWait(focusLow);
                int i;
                for (i = 0; i < 90; i++) {

                        temp = receivePacketFromCam();
                        if (max < temp) {
                                max = temp;
                                maxi = (int) (i * 1.2);
                        }

                }

                setFocusToMm(12 + 108 - maxi);
        }

        else {
                setFocusToMm(focusLow);
                setFocusToMmNoWait(focusHigh);
                int i;
                for (i = 0; i < 90; i++) {

                        temp = receivePacketFromCam();
                        if (max < temp) {
                                max = temp;
                                maxi = (int) (i * 1.2);
                        }

                }

                setFocusToMm(12 + maxi);
        }
        // 7,225sec for move focusrange
        //7,1sec --- 180= 5 *36
        //108 /(180/2)  =1,2
        //7,225/109 =0,066

}

/*
 * creates UDP socket on local Port
 */
void createsUDPSocketOnPort(int port) {
        sockCam = socket(AF_INET, SOCK_DGRAM, 0);
        if (sockCam < 0)
                error("socketCAM");

        serverCam.sin_family = AF_INET;
        serverCam.sin_addr.s_addr = INADDR_ANY;
        serverCam.sin_port = htons(port);

        bind(sockCam, (struct sockaddr *) &serverCam, sizeof(serverCam));
}

/*
 * read a packet from camera
 */
long receivePacketFromCam() {

        createsUDPSocketOnPort(1234);

        char byteArray[8];
        char temp[8] = { 0 };
        long tempArray[5];
        int i;
        for (i = 0; i < 2; i++) {

                n = recvfrom(sockCam, byteArray, 8, 0, (struct sockaddr *) &from,
                                &length);

                memcpy(temp, &byteArray, sizeof(long));
                tempArray[i] = *(long *) temp;

                if (n < 0)
                        error("recvfrom");
        }
        long ret = (long) ((tempArray[0] + tempArray[1]) / 2);
        close(sockCam);

        return ret;
}

void movePtu(PTUCOMMAND command, int timesToSend) {

        switch (command) {

        case RIGHT:
                sendCommand("40");
                sprintf(commandString, "%d", timesToSend);
                sendCommand(commandString);
                receivePacket();
                break;
        case LEFT:
                sendCommand("41");
                sprintf(commandString, "%d", timesToSend);
                sendCommand(commandString);
                receivePacket();
                break;
        case UP:
                sendCommand("42");
                sprintf(commandString, "%d", timesToSend);
                sendCommand(commandString);
                receivePacket();
                break;
        case DOWN:
                sendCommand("43");
                sprintf(commandString, "%d", timesToSend);
                sendCommand(commandString);
                receivePacket();
                break;
        }

}

void Initialization_A(void) {
        movePtu(LEFT, 360);
        movePtu(DOWN, 95);
        movePtu(RIGHT, 178);
        movePtu(UP, 41);
}
