/* GStreamer
 * Copyright (C) 2006 David A. Schleef <ds@schleef.org>
 *
 * gstinotifysrc.c:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/**
 * SECTION:element-inotifysrc
 * @see_also: #GstFileSrc
 *
 * Reads buffers from sequentially named files. If used together with an image
 * decoder, one needs to use the #Gstinotifysrc:caps property or a capsfilter
 * to force to caps containing a framerate. Otherwise image decoders send EOS
 * after the first picture.
 *
 * File names are created by replacing "%%d" with the index using printf().
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch inotifysrc location="img.%04d.png" index=0 caps="image/png,framerate=\(fraction\)12/1" ! \
 *     pngdec ! ffmpegcolorspace ! theoraenc ! oggmux ! \
 *     filesink location="images.ogg"
 * ]| This pipeline creates a video file "images.ogg" by joining multiple PNG
 * files named img.0000.png, img.0001.png, etc.
 * </refsect2>
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>

//#include <sys/inotify.h>
#include <sys/select.h>

#include <sys/types.h>
#include  <signal.h>
#include <sys/time.h>
#include <unistd.h>

#include <linux/inotify.h>
#include "gstinotifysrc.h"


static GstFlowReturn gst_inotifysrc_create (GstPushSrc * src,
    GstBuffer ** buffer);

static void gst_inotifysrc_dispose (GObject * object);

static void gst_inotifysrc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_inotifysrc_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstCaps *gst_inotifysrc_getcaps (GstBaseSrc * src);
static gboolean gst_inotifysrc_query (GstBaseSrc * src, GstQuery * query);
static gboolean gst_inotifysrc_stop (GstBaseSrc * src);
static gboolean gst_inotifysrc_start (GstBaseSrc * src);
static GstStaticPadTemplate gst_inotifysrc_pad_template =
GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY);

GST_DEBUG_CATEGORY_STATIC (gst_inotifysrc_debug);
#define GST_CAT_DEFAULT gst_inotifysrc_debug

static const GstElementDetails gst_inotifysrc_details =
GST_ELEMENT_DETAILS ("Multi-File Source",
    "Source/File",
    "Read a sequentially named set of files into buffers",
    "David Schleef <ds@schleef.org>");

enum
{
  ARG_0,
  ARG_LOCATION,
  ARG_INDEX,
  ARG_CAPS,
  ARG_TIMEOUT,
  ARG_SLEEP,
  ARG_COMMANDONCE,
  ARG_COMMANDEACH,
  ARG_EXCLUDESTR,
  ARG_SIGUSR,
  ARG_DELETEFILE

};

#define DEFAULT_LOCATION "%05d"
#define DEFAULT_INDEX 0



GST_BOILERPLATE (Gstinotifysrc, gst_inotifysrc, GstPushSrc,
    GST_TYPE_PUSH_SRC);

static void
gst_inotifysrc_base_init (gpointer g_class)
{
  GstElementClass *gstelement_class = GST_ELEMENT_CLASS (g_class);

  GST_DEBUG_CATEGORY_INIT (gst_inotifysrc_debug, "inotifysrc", 0,
      "inotifysrc element");

  gst_element_class_add_pad_template (gstelement_class,
      gst_static_pad_template_get (&gst_inotifysrc_pad_template));
  gst_element_class_set_details (gstelement_class, &gst_inotifysrc_details);
}

static void
gst_inotifysrc_class_init (GstinotifysrcClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstPushSrcClass *gstpushsrc_class = GST_PUSH_SRC_CLASS (klass);
  GstBaseSrcClass *gstbasesrc_class = GST_BASE_SRC_CLASS (klass);

  gobject_class->set_property = gst_inotifysrc_set_property;
  gobject_class->get_property = gst_inotifysrc_get_property;

  g_object_class_install_property (gobject_class, ARG_LOCATION,
      g_param_spec_string ("location", "File Location",
          "Pattern to create file names of input files.  File names are "
          "created by calling sprintf() with the pattern and the current "
          "index.", DEFAULT_LOCATION, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_INDEX,
      g_param_spec_int ("index", "File Index",
          "Index to use with location property to create file names.  The "
          "index is incremented by one for each buffer read.",
          0, INT_MAX, DEFAULT_INDEX, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_CAPS,
      g_param_spec_boxed ("caps", "Caps",
          "Caps describing the format of the data.",
          GST_TYPE_CAPS, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_TIMEOUT,
      g_param_spec_int ("timeout", "Timeout",
          "Timeout to wait for new file in ms (-1 for none)",
          -1, INT_MAX, -1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_SLEEP,
      g_param_spec_int ("sleep", "Sleep",
          "Sleep ms before generating each output",
          -1, INT_MAX, -1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_COMMANDONCE,
      g_param_spec_string ("commandonce", "Commandonce",
          "This command will run detached and "
	  "generate outputfiles continuously", NULL, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_COMMANDEACH,
      g_param_spec_string ("commandeach", "Commandeach",
          "This command will be called each "
	  "time a new buffer shall be read", NULL, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_EXCLUDESTR,
      g_param_spec_string ("excludestr", "Excludestr",
          "Ignore filenames with this substring ", "~", G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_SIGUSR,
      g_param_spec_boolean ("sigusr", "Sigusr", "Wether we send SIGUSR1 to the child"
	  "we started with commandonce",
          FALSE, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_DELETEFILE,
      g_param_spec_boolean ("delete", "Delete", "Wether we should delete files after read",
          TRUE, (GParamFlags)G_PARAM_READWRITE));


  gobject_class->dispose = gst_inotifysrc_dispose;

  gstbasesrc_class->get_caps = gst_inotifysrc_getcaps;
  gstbasesrc_class->query = gst_inotifysrc_query;
  gstbasesrc_class->start = gst_inotifysrc_start;
  gstbasesrc_class->stop = gst_inotifysrc_stop;
  gstpushsrc_class->create = gst_inotifysrc_create;

  if (sizeof (off_t) < 8) {
    GST_LOG ("No large file support, sizeof (off_t) = %" G_GSIZE_FORMAT,
        sizeof (off_t));
  }
}

static void
gst_inotifysrc_init (Gstinotifysrc * inotifysrc,
    GstinotifysrcClass * g_class)
{
  inotifysrc->index = DEFAULT_INDEX;
  inotifysrc->dirname = g_strdup (DEFAULT_LOCATION);
  inotifysrc->successful_read = FALSE;

  inotifysrc->commandeach = NULL;
  inotifysrc->commandonce = NULL;
  inotifysrc->excludestr = "~";
  inotifysrc->timeout=-1;
  inotifysrc->sleep=-1;
  inotifysrc->deletefile=TRUE;
}

static void
gst_inotifysrc_dispose (GObject * object)
{
  Gstinotifysrc *src = GST_INOTIFYSRC (object);

  g_free (src->dirname);
  src->dirname = NULL;
  if (src->caps)
    gst_caps_unref (src->caps);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static GstCaps *
gst_inotifysrc_getcaps (GstBaseSrc * src)
{
  Gstinotifysrc *multi_file_src = GST_INOTIFYSRC (src);

  GST_DEBUG_OBJECT (src, "returning %" GST_PTR_FORMAT, multi_file_src->caps);

  if (multi_file_src->caps) {
    return gst_caps_ref (multi_file_src->caps);
  } else {
    return gst_caps_new_any ();
  }
}

static gboolean gst_inotifysrc_start (GstBaseSrc * src){

  Gstinotifysrc *insrc;

  insrc = GST_INOTIFYSRC (src);
  insrc->i=0;
  insrc->fd = inotify_init();
  if ( insrc->fd < 0 )
  {
    printf("inotify init failed with error # %d\n", insrc->fd);
    return FALSE;
  }
  insrc->wd = inotify_add_watch( insrc->fd,insrc->dirname , IN_CLOSE_WRITE );

  FD_ZERO(&(insrc->rfds));
  FD_SET(insrc->fd, &(insrc->rfds));

  
  g_print("Finished addinf Directory to watchlist\n");
  if (insrc->commandonce){
    pid_t pid = 0;
    pid = fork();
    if (pid == 0) {
	//usleep(500000);
        execl("/bin/sh","-c",insrc->commandonce,NULL);
        //system(insrc->commandonce);
	exit(0);
        }
    //usleep(500000);    
    g_print("Child process has PID: %d\n",pid);
    insrc->childpid=pid;
  }
  return TRUE;
}

static gboolean gst_inotifysrc_stop (GstBaseSrc * src){
  Gstinotifysrc *insrc;
  g_print("Stop called\n");
  insrc = GST_INOTIFYSRC (src);
   
  inotify_rm_watch( insrc->fd,insrc-> wd );
  close( insrc->fd );
  if (insrc->commandonce){
    g_print("sending kill to %d\n",insrc->childpid);
    kill(insrc->childpid,SIGKILL);
    //kill(insrc->childpid+1,SIGKILL);
    //kill(insrc->childpid+2,SIGKILL);
  }
  return TRUE;
}

static gboolean
gst_inotifysrc_query (GstBaseSrc * src, GstQuery * query)
{
  gboolean res;
  Gstinotifysrc *mfsrc;

  mfsrc = GST_INOTIFYSRC (src);

  switch (GST_QUERY_TYPE (query)) {
    case GST_QUERY_POSITION:
    {
      GstFormat format;

      gst_query_parse_position (query, &format, NULL);
      switch (format) {
        case GST_FORMAT_BUFFERS:
        case GST_FORMAT_DEFAULT:
          gst_query_set_position (query, GST_FORMAT_BUFFERS, mfsrc->index);
          res = TRUE;
          break;
        default:
          res = GST_BASE_SRC_CLASS (parent_class)->query (src, query);
          break;
      }
      break;
    }
    default:
      res = GST_BASE_SRC_CLASS (parent_class)->query (src, query);
      break;
  }
  return res;
}




static gboolean
gst_inotifysrc_set_location (Gstinotifysrc * src, const gchar * location)
{
  g_free (src->dirname);
  if (location != NULL) {
    src->dirname = g_strdup (location);
  } else {
    src->dirname = NULL;
  }

  return TRUE;
}

static void
gst_inotifysrc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstinotifysrc *src = GST_INOTIFYSRC (object);

  switch (prop_id) {
    case ARG_LOCATION:
      gst_inotifysrc_set_location (src, g_value_get_string (value));
      break;
    case ARG_INDEX:
      src->index = g_value_get_int (value);
      break;
    case ARG_CAPS:
    {
      const GstCaps *caps = gst_value_get_caps (value);
      GstCaps *new_caps;

      if (caps == NULL) {
        new_caps = gst_caps_new_any ();
      } else {
        new_caps = gst_caps_copy (caps);
      }
      gst_caps_replace (&src->caps, new_caps);
      gst_pad_set_caps (GST_BASE_SRC_PAD (src), new_caps);
    }
      break;
    case ARG_TIMEOUT:
      src->timeout = g_value_get_int (value);
      break;
    case ARG_SLEEP:
      src->sleep = g_value_get_int (value);
      break;
    case ARG_COMMANDONCE:
      src->commandonce=g_strdup(g_value_get_string (value));
      break;
    case ARG_COMMANDEACH:
      src->commandeach=g_strdup(g_value_get_string (value));
      break;
    case ARG_EXCLUDESTR:
      src->excludestr=g_strdup(g_value_get_string (value));
      break;
    case ARG_SIGUSR:
      src->sigusr=g_value_get_boolean (value);
      break;
    case ARG_DELETEFILE:
      src->deletefile=g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_inotifysrc_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstinotifysrc *src = GST_INOTIFYSRC (object);

  switch (prop_id) {
    case ARG_LOCATION:
      g_value_set_string (value, src->dirname);
      break;
    case ARG_INDEX:
      g_value_set_int (value, src->index);
      break;
    case ARG_CAPS:
      gst_value_set_caps (value, src->caps);
      break;
    case ARG_TIMEOUT:
      g_value_set_int (value, src->timeout);
      break;
    case ARG_SLEEP:
      g_value_set_int (value, src->sleep);
      break;
    case ARG_COMMANDONCE:
      g_value_set_string (value, src->commandonce);
      break;
    case ARG_COMMANDEACH:
      g_value_set_string (value, src->commandeach);
      break;
    case ARG_EXCLUDESTR:
      g_value_set_string (value, src->excludestr);
      break;
    case ARG_SIGUSR:
      g_value_set_boolean (value, src->sigusr);
      break;
    case ARG_DELETEFILE:
      g_value_set_boolean (value, src->deletefile);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static gchar *
gst_inotifysrc_get_filename (Gstinotifysrc * inotifysrc)
{
  gchar *filename;

  filename = g_strdup_printf (inotifysrc->dirname, inotifysrc->index);

  return filename;
}

static GstFlowReturn
gst_inotifysrc_create (GstPushSrc * src, GstBuffer ** buffer)
{
  Gstinotifysrc *insrc;
  gsize size;
  gchar *data;
  gchar *filename;
  GstBuffer *buf;
  gboolean ret;
  GError *error = NULL;

  insrc = GST_INOTIFYSRC (src);
  int result;

  if (insrc->commandeach){
	g_print("Generating each output\n");
	if (insrc->sleep>0)usleep(1000*insrc->sleep);
	result=system(insrc->commandeach);
	if (result==-1) return GST_FLOW_UNEXPECTED;
  }
  if ((insrc->commandonce)&&(insrc->sigusr)){
	//TODO send signal to thread    
	kill(insrc->childpid,SIGUSR1);
  }


  int length,i=0;

  if (insrc->i==0){
    g_print("Wait\n");
    if (insrc->timeout>0){
      insrc->tv.tv_sec = insrc->timeout/1000;
      insrc->tv.tv_usec = (insrc->timeout%1000)*1000;
      if(select(insrc->fd+1, &(insrc->rfds), NULL, NULL, &(insrc->tv))>0) insrc->length = read( insrc->fd, insrc->buffer, EVENT_BUF_LEN );
      else return GST_FLOW_ERROR;
    }
    else insrc->length = read( insrc->fd, insrc->buffer, EVENT_BUF_LEN );
    g_print("Finished wait, read %d bytes of events\n",insrc->length);
  }  
  //get filename from event buffer
  if (insrc->i<insrc->length){
    struct inotify_event *pevent    = (struct inotify_event *) &insrc->buffer[insrc->i];
    g_print("Event NR. %d, Filename: %s\n",insrc->i,pevent->name); 
    insrc->i += sizeof(struct inotify_event) + pevent->len;
    strncpy(insrc->filename,insrc->dirname , 512);
    if (insrc->dirname[strlen(insrc->dirname)-1]!='/') strncat(insrc->filename,"/",512);
    strncat(insrc->filename,pevent->name,512);
    g_print("Complete Name: %s\n",insrc->filename);
  }
  if (insrc->i>=insrc->length)insrc->i=0;//to make it start again
/*
  while (insrc->i<insrc->length){
  	struct inotify_event *pevent    = (struct inotify_event *) &insrc->buffer[i];
	g_print("Event NR. %d, Filename: %s\n",i,pevent->name); 
	i += sizeof(struct inotify_event) + pevent->len;
  }*/

  //g_print(insrc->buffer);
  //filename = gst_inotifysrc_get_filename (insrc);

  GST_DEBUG_OBJECT (insrc, "reading from file \"%s\".", insrc->filename);

  ret = g_file_get_contents (insrc->filename, &data, &size, &error);
  if (!ret) {
    if (insrc->successful_read) {
      /* If we've read at least one buffer successfully, not finding the
       * next file is EOS. */
      g_free (filename);
      if (error != NULL)
        g_error_free (error);
      return GST_FLOW_UNEXPECTED;
    } else {
      goto handle_error;
    }
  }

  insrc->successful_read = TRUE;
  insrc->index++;

  buf = gst_buffer_new ();
  GST_BUFFER_DATA (buf) = (unsigned char *) data;
  GST_BUFFER_MALLOCDATA (buf) = GST_BUFFER_DATA (buf);
  GST_BUFFER_SIZE (buf) = size;
  GST_BUFFER_OFFSET (buf) = insrc->offset;
  GST_BUFFER_OFFSET_END (buf) = insrc->offset + size;
  insrc->offset += size;
  gst_buffer_set_caps (buf, insrc->caps);

  GST_DEBUG_OBJECT (insrc, "read file \"%s\".", filename);

  g_free (filename);
  if (insrc->deletefile)g_remove(insrc->filename);
  *buffer = buf;
  return GST_FLOW_OK;

handle_error:
  {
    if (error != NULL) {
      GST_ELEMENT_ERROR (insrc, RESOURCE, READ,
          ("Error while reading from file \"%s\".", filename),
          ("%s", error->message));
      g_error_free (error);
    } else {
      GST_ELEMENT_ERROR (insrc, RESOURCE, READ,
          ("Error while reading from file \"%s\".", filename),
          ("%s", g_strerror (errno)));
    }
    g_free (filename);
    return GST_FLOW_ERROR;
  }
}
static gboolean init_inotifysrc (GstPlugin * plugin){
return gst_element_register (plugin, "inotifysrc", GST_RANK_NONE, GST_TYPE_INOTIFYSRC);}
GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,GST_VERSION_MINOR,"inotifysrc","Template inotifysrc",init_inotifysrc,VERSION,"LGPL","GStreamer","http:gstreamer.net")
