/*
 * arduinolib.c
 *
 *  Created on: 04.11.2011
 *      Author: Johannes Liebrecht
 */
#include<math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <assert.h>
#include <time.h>
#include <sys/time.h>
//#include "arduinolib.h"

#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */

enum _PTUCOMMAND {
	LEFT, RIGHT, UP, DOWN
};

typedef enum _PTUCOMMAND PTUCOMMAND;

void setZoomToMm(int desiredValue);
void setFocusToMm(int desiredValue);
void createConnectionToArduino(const char ip[], int port);
void disconnectConnectionToArduino(void);
int* getAnalogValues(void);
void autoFocus();

int receivePacket(void);
void sendCommand(const char command[]);
long receivePacketFromCam();

void movePtu(PTUCOMMAND, int timesToSend);
void Initialization_A(void);
void Initialization_Lens(int objectiveMin, int objectiveMax);
int openSerialPort(char* string);
void writeSerialPort(char* string);
void readSerialPort();
void initializiationLookUpTable(void);
void defaultAutofokus(void);

#define UDP_RX_PACKET_MAX_SIZE 24
#define SET_FOCUS_MIN	1
#define SET_FOCUS_MAX	100

int _focusMin = 173, _focusMax = 723, _zoomMin = 172, _zoomMax = 727,
		_objectiveMin = 12, _objectiveMax = 120;
long _focusTimeLowToHiGH, _focusTimeHighToLow, _zoomTimeLowToHigh,
		_zoomTimeHighToLow;
int sock, n, sockCam;
unsigned int length;
struct sockaddr_in server, from, serverCam, serverFrom;
char commandString[10];
char receiveBuffer[UDP_RX_PACKET_MAX_SIZE];
int analogInputValue[2];
char LastReceivedPTUBuffer[100];

int _fd; /* File descriptor for the serial port */

void error(const char *);

/*
 * send command to arduino
 */

void sendCommand(const char command[]) {

	n = sendto(sock, command, strlen(command), 0,
			(const struct sockaddr *) &server, length);
	if (n < 0)
		error("Sendto");
}



/*
 * wait for a packet from arduino
 * return : received packet
 */

int receivePacket() {

	n = recvfrom(sock, receiveBuffer, UDP_RX_PACKET_MAX_SIZE, 0,
			(struct sockaddr *) &from, &length);
	if (n < 0)
		error("recvfrom");

	return atoi(receiveBuffer);
}

/*
 * get the current zoom and focus values from arduino
 */
int* getAnalogValues() {
	sendCommand("00000000");
	analogInputValue[0] = receivePacket();
	analogInputValue[1] = receivePacket();

	//printf("%d,%d", analogInputValue[0], analogInputValue[1]);
	return analogInputValue;
}

/*
 * creates UDP socket
 * first : arduino ip
 * second : arduino port
 */
void createConnectionToArduino(const char ip[], int port) {
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0)
		error("socket");

	server.sin_family = AF_INET;
	server.sin_addr.s_addr = inet_addr(ip);
	server.sin_port = htons(port);
	length = sizeof(struct sockaddr_in);
}

/*
 * close UDP socket
 */
void disconnectConnectionToArduino() {
	close(sock);
}

/*
 * set objective zoom and wait until set
 * desiredValue : objectiveMin - objectiveMax
 */
void setZoomToMm(int desiredValue) {
	assert(desiredValue>=_objectiveMin);
	assert(desiredValue<=_objectiveMax);
	sendCommand("00000001");
	int temp;
	double temp2;
	temp2 = (double) (_zoomMax - _zoomMin) / (double) (_objectiveMax
			- _objectiveMin);

	temp = (int) (((double) desiredValue * temp2) + ((double) _zoomMin
			- (double) _objectiveMin * temp2));

	sprintf(commandString, "%d", temp);
	sendCommand(commandString);
	receivePacket();
}

void setZoomToMmNoWait(int desiredValue) {
	assert(desiredValue>=_objectiveMin);
	assert(desiredValue<=_objectiveMax);
	sendCommand("00000001");
	int temp;
	double temp2;
	temp2 = (double) (_zoomMax - _zoomMin) / (double) (_objectiveMax
			- _objectiveMin);

	temp = (int) (((double) desiredValue * temp2) + ((double) _zoomMin
			- (double) _objectiveMin * temp2));

	sprintf(commandString, "%d", temp);
	sendCommand(commandString);
	//receivePacket();
}


/*
 * set objective focus and wait until set
 * desiredValue : 1-100
 */
void setFocusToMm(int desiredValue) {
	assert(desiredValue>=SET_FOCUS_MIN);
	assert(desiredValue<=SET_FOCUS_MAX);
	sendCommand("00000010");
	int temp;
	temp = (int) (((_focusMax - _focusMin) * desiredValue) / 100);
	sprintf(commandString, "%d", temp + _focusMin);
	sendCommand(commandString);
	receivePacket();
}

/*
 * set objective focus
 * !!Attention!! did not wait until objective focus is set
 * desiredValue : 1-100
 */
void setFocusToMmNoWait(int desiredValue) {
	assert(desiredValue>=SET_FOCUS_MIN);
	assert(desiredValue<=SET_FOCUS_MAX);
	sendCommand("00000011");
	int temp;
	temp = (int) ((_focusMax - _focusMin) * desiredValue / 100);

	sprintf(commandString, "%d", temp + _focusMin);
	sendCommand(commandString);
}

void error(const char *msg) {
	perror(msg);
	exit(0);
}

void autoFocus() {
	long max = 0;
	long temp = 0;
	long maxi = 0;

	struct timeval start, end;
	long time;
	gettimeofday(&start, NULL);
	receivePacketFromCam();
	receivePacketFromCam();
	receivePacketFromCam();
	receivePacketFromCam();
	receivePacketFromCam();
	receivePacketFromCam();
	receivePacketFromCam();
	receivePacketFromCam();
	receivePacketFromCam();
	receivePacketFromCam();

	gettimeofday(&end, NULL);
	time = (end.tv_sec * 1000000 + end.tv_usec) - (start.tv_sec * 1000000
			+ start.tv_usec);

	getAnalogValues();

	if (analogInputValue[0] > ((int) (_focusMax - _focusMin) / 2) + _focusMin) {

		int iterate;
		iterate = (int) ((long) _focusTimeHighToLow / (time / 10));

		setFocusToMm(SET_FOCUS_MAX);

		setFocusToMmNoWait(SET_FOCUS_MIN);
		int i;
		//test
		FILE *data;
		data = fopen("autofokus.txt", "a");
		if (data != NULL) {

			for (i = 0; i < iterate; i++) {
				temp = receivePacketFromCam();

				if (max < temp) {
					max = temp;
					maxi = (long) round((double) i * ((double) SET_FOCUS_MAX
							/ (double) iterate));
				}

				//test
				fwrite(&temp, sizeof(long), 1, data);
			}
			int x = 11111111, y =120;
			fwrite(&y, sizeof(long), 1, data);
			fwrite(&maxi, sizeof(long), 1, data);
			fwrite(&x, sizeof(long), 1, data);

			fclose(data);
		}

		setFocusToMm(SET_FOCUS_MAX - maxi);
		printf("TestHTLFocus:%ld", (SET_FOCUS_MAX - maxi));

	}

	else {
		int iterate;
		iterate = (int) ((long) _focusTimeLowToHiGH / (time / 10));

		setFocusToMm(SET_FOCUS_MIN);

		setFocusToMmNoWait(SET_FOCUS_MAX);
		int i;
		//test
		FILE *data;
		data = fopen("autofokus.txt", "a");
		if (data != NULL) {

			for (i = 0; i < iterate; i++) {
				temp = receivePacketFromCam();

				if (max < temp) {
					max = temp;
					maxi = (long) round(((double) i * ((double) SET_FOCUS_MAX
							/ (double) iterate)));
				}

				//test
				fwrite(&temp, sizeof(long), 1, data);
			}
			int x = 11111111, y =12;
			fwrite(&maxi, sizeof(long), 1, data);
			fwrite(&y, sizeof(long), 1, data);
			fwrite(&x, sizeof(long), 1, data);

			fclose(data);
		}

		setFocusToMm(SET_FOCUS_MIN + maxi);
		printf("TestLTHFocus:%ld", SET_FOCUS_MIN + maxi);

	}
}

/*
 * creates UDP socket on local Port
 */
void createsUDPSocketOnPort(int port) {
	sockCam = socket(AF_INET, SOCK_DGRAM, 0);
	if (sockCam < 0)
		error("socketCAM");

	serverCam.sin_family = AF_INET;
	serverCam.sin_addr.s_addr = INADDR_ANY;
	serverCam.sin_port = htons(port);

	bind(sockCam, (struct sockaddr *) &serverCam, sizeof(serverCam));
}

/*
 * read a packet from camera
 */
long receivePacketFromCam() {

	createsUDPSocketOnPort(1234);

	char byteArray[8];
	char temp[8] = { 0 };
	long ret;

	n = recvfrom(sockCam, byteArray, 8, 0, (struct sockaddr *) &from, &length);

	memcpy(temp, &byteArray, sizeof(long));
	ret = *(long *) temp;

	if (n < 0)
		error("recvfrom");

	close(sockCam);

	return ret;
}
/*
 * Move PTU YAAN
 */
void movePtu(PTUCOMMAND command, int timesToSend) {

	switch (command) {

	case RIGHT:
		sendCommand("00000100");
		sprintf(commandString, "%d", timesToSend);
		sendCommand(commandString);
		receivePacket();
		break;
	case LEFT:
		sendCommand("00000101");
		sprintf(commandString, "%d", timesToSend);
		sendCommand(commandString);
		receivePacket();
		break;
	case UP:
		sendCommand("00000110");
		sprintf(commandString, "%d", timesToSend);
		sendCommand(commandString);
		receivePacket();
		break;
	case DOWN:
		sendCommand("00000111");
		sprintf(commandString, "%d", timesToSend);
		sendCommand(commandString);
		receivePacket();
		break;
	}

}

/*
 * initialization for PTU Yaan
 */
void Initialization_A(void) {
	movePtu(LEFT, 360);
	movePtu(DOWN, 95);
	movePtu(RIGHT, 178);
	movePtu(UP, 41);
}
// initialization camera objective
void Initialization_Lens(int objectiveMin, int objectiveMax) {
	_objectiveMin = objectiveMin;
	_objectiveMax = objectiveMax;

	//get focus/zoom values
	sendCommand("00001000");

	_focusMax = receivePacket();
	_focusMin = receivePacket();
	_zoomMin = receivePacket();
	_zoomMax = receivePacket();

	//send to arduino
	sendCommand("00001001");
	receivePacket();

	sprintf(commandString, "%d", _focusMin);
	sendCommand(commandString);
	receivePacket();

	sprintf(commandString, "%d", _focusMax);
	sendCommand(commandString);
	receivePacket();

	sprintf(commandString, "%d", _zoomMin);
	sendCommand(commandString);
	receivePacket();

	sprintf(commandString, "%d", _zoomMax);
	sendCommand(commandString);
	receivePacket();

	sprintf(commandString, "%d", _objectiveMin);
	sendCommand(commandString);
	receivePacket();

	sprintf(commandString, "%d", _objectiveMax);
	sendCommand(commandString);
	receivePacket();

	setFocusToMm(1);
	struct timeval start, end;
	gettimeofday(&start, NULL);
	setFocusToMm(100);
	gettimeofday(&end, NULL);
	_focusTimeLowToHiGH = (end.tv_sec * 1000000 + end.tv_usec) - (start.tv_sec
			* 1000000 + start.tv_usec);

	gettimeofday(&start, NULL);
	setFocusToMm(1);
	gettimeofday(&end, NULL);
	_focusTimeHighToLow = (end.tv_sec * 1000000 + end.tv_usec) - (start.tv_sec
			* 1000000 + start.tv_usec);

	setZoomToMm(_objectiveMin);
	gettimeofday(&start, NULL);
	setZoomToMm(_objectiveMax);
	gettimeofday(&end, NULL);
	_zoomTimeLowToHigh = (end.tv_sec * 1000000 + end.tv_usec) - (start.tv_sec
			* 1000000 + start.tv_usec);

	gettimeofday(&start, NULL);
	setZoomToMm(_objectiveMin);
	gettimeofday(&end, NULL);
	_zoomTimeHighToLow = (end.tv_sec * 1000000 + end.tv_usec) - (start.tv_sec
			* 1000000 + start.tv_usec);

	//save to data
	FILE *data;
	data = fopen("config.txt", "w");
	if (data != NULL) {
		fwrite(&_focusMin, sizeof(int), 1, data);
		fwrite(&_focusMax, sizeof(int), 1, data);
		fwrite(&_zoomMin, sizeof(int), 1, data);
		fwrite(&_zoomMax, sizeof(int), 1, data);
		fwrite(&_objectiveMin, sizeof(int), 1, data);
		fwrite(&_objectiveMax, sizeof(int), 1, data);
		fwrite(&_focusTimeLowToHiGH, sizeof(long), 1, data);
		fwrite(&_focusTimeHighToLow, sizeof(long), 1, data);
		fwrite(&_zoomTimeLowToHigh, sizeof(long), 1, data);
		fwrite(&_zoomTimeHighToLow, sizeof(long), 1, data);

		fclose(data);
	}
}
/*
 *  read last saved _focusmin,_focusmax,_zoomMin,_zoomMax,_objectiveMin,
 * _objectiveMax,_focusTimeLowToHigh,_focusTimeHighToLow,_zoomTimeHighToLow and_zoomTimeLowToHigh
 *  from file and send to Arduino for initialization.
 */
void Initialization_Arduino() {
	//read data
	FILE *data;
	data = fopen("config.txt", "r");
	if (data != NULL) {
		fread(&_focusMin, sizeof(int), 1, data);
		// printf("%d\n",_focusMin);
		fread(&_focusMax, sizeof(int), 1, data);
		// printf("%d\n",_focusMax);
		fread(&_zoomMin, sizeof(int), 1, data);
		// printf("%d\n",_zoomMin);
		fread(&_zoomMax, sizeof(int), 1, data);
		// printf("%d\n",_zoomMax);
		fread(&_objectiveMin, sizeof(int), 1, data);
		fread(&_objectiveMax, sizeof(int), 1, data);
		fread(&_focusTimeLowToHiGH, sizeof(long), 1, data);
		// printf("%ld\n",_focusTimeLowToHiGH);
		fread(&_focusTimeHighToLow, sizeof(long), 1, data);
		//printf("%ld\n",_focusTimeHighToLow);
		fread(&_zoomTimeLowToHigh, sizeof(long), 1, data);
		//printf("%ld\n",_zoomTimeLowToHigh);
		fread(&_zoomTimeHighToLow, sizeof(long), 1, data);
		//printf("%ld\n",_zoomTimeHighToLow);

		fclose(data);

		//send to arduino
		sendCommand("00001001");
		receivePacket();

		sprintf(commandString, "%d", _focusMin);
		sendCommand(commandString);
		receivePacket();

		sprintf(commandString, "%d", _focusMax);
		sendCommand(commandString);
		receivePacket();

		sprintf(commandString, "%d", _zoomMin);
		sendCommand(commandString);
		receivePacket();

		sprintf(commandString, "%d", _zoomMax);
		sendCommand(commandString);
		receivePacket();

		sprintf(commandString, "%d", _objectiveMin);
		sendCommand(commandString);
		receivePacket();

		sprintf(commandString, "%d", _objectiveMax);
		sendCommand(commandString);
		receivePacket();

	}
}

/*
 * Open serial port (e.g. "/dev/ttyArduino")
 * Before you need to create a process with socat
 * sudo socat PTY,link=/dev/ttyArduino,raw,echo=0, UDP4:134.100.13.161:8888
 * and give rights with sudo chmod o+rw /dev/ttyArduino
 * Returns the file descriptor on success or -1 on error.
 */

int openSerialPort(char * string) {

	_fd = open(string, O_RDWR | O_NOCTTY | O_NDELAY);
	//		_fd = open("/dev/ttyUSB0", O_RDWR | O_NOCTTY | O_NDELAY);

	if (_fd == -1) {
		perror("open_port: Unable to open /dev/ttyArduino - ");
	} else
		//		fcntl(_fd, F_SETFL, 0); //blocking read
		writeSerialPort("00001010"); //Arduino remember Port
	sleep(1);
	return (_fd);
}
/*
 * write to Virtual Serial Port
 */
void writeSerialPort(char* string) {
	int n = write(_fd, string, strlen(string));
	if (n < 0)
		fputs("write() of bytes failed!\n", stderr);
}

/*
 * read from virtual Serial Port
 */
void readSerialPort() {
	fcntl(_fd, F_SETFL, FNDELAY); //read no blocking
	char buf[100];

	int n = read(_fd, buf, sizeof buf);
	if (n > 0) {
		buf[n] = 0;
		printf("%s\n", buf);

		int i;
		for (i = 0; i < 100; i++) {
			LastReceivedPTUBuffer[i] = buf[i];
		}
	}
}
/*
 * reset PTU C46
 */
void resetPTU(void) {
	writeSerialPort("r ");
}

/*
 * set PTU pan and tilt
 * pan : from -180 to +180
 * tilt: from -90 to +90
 */
void setPTUPanTilt(double pan, double tilt) {
	double minimumPanPosition = -3090;
	double maximumPanPosition = 3090;
	double minimumTiltPosition = -905;
	double maximumTiltPosition = 603;
	int panValue = 0;
	int tiltValue = 0;

	if (pan < 0) {
		panValue = (int) round((minimumPanPosition / (double) (-180) * pan));
	} else {
		panValue = (int) round((maximumPanPosition / (double) (180) * pan));
	}

	if (tilt < 0) {
		tiltValue = (int) round((minimumTiltPosition / (double) (-90) * tilt));

	} else {
		tiltValue = (int) round((maximumTiltPosition / (double) (90) * tilt));

	}
	char temp[10] = { 0 };
	char temp2[10] = { 0 };
	sprintf(temp, "%d", panValue);
	sprintf(temp2, "%d", tiltValue);

	char text[100] = { 0 };
	strcat(text, "s pp");
	strcat(text, temp);
	strcat(text, " tp");
	strcat(text, temp2);
	strcat(text, " a ");

	writeSerialPort(text);
}

/*
 * initialization the LookUpTable with zoom = _objectiveMin,_objectiveMin+5,_objectiveMin+10,...
 */
void initializiationLookUpTable(void) {
	FILE *data;
	data = fopen("LookUpTable.txt", "w");
	if (data != NULL) {
		int i;
		for (i = 0; i <= (int) ((_objectiveMax - _objectiveMin) / 5); i++) {
			if ((_objectiveMin + i * 5) <= _objectiveMax) {
				setZoomToMm(_objectiveMin + i * 5);
				autoFocus();
				getAnalogValues();
				//fwrite(&analogInputValue[0], sizeof(int), 1, data);
				fprintf(data,"%d\n",analogInputValue[0]);
			}
		}
		fclose(data);
	}
}

/*
 * reads the zoom and set focus from LookUpTable
 */


int setFocusFromLUT(void) {
	FILE *data;
	data = fopen("LookUpTable.txt", "r");
	ssize_t getlineresult=0;
        int setfocusto=-1;
	//printf("Hello World\n");
	//char intchar[30];
	char* intchar;
	size_t bytetoread=0;
	if (data != NULL) {
		int i;
		int array[(int) ((_objectiveMax - _objectiveMin) / 5) + 1];
		for (i = 0; i <= (int) ((_objectiveMax - _objectiveMin) / 5); i++) {
			//printf("Hello World\n");
			//fread(&array[i], sizeof(int), 1, data);
			getlineresult=getline(&intchar,&bytetoread,data);
			//printf("getlineresult %d --\n",getlineresult);
			array[i]=atoi(intchar);
			//printf("%d --",array[i]);
			//printf("%s --\n",intchar);

		}
		getAnalogValues();

		int zoom;
		double temp;
		temp = (double) (_zoomMax - _zoomMin) / (double) (_objectiveMax
				- _objectiveMin);
		zoom = (int) round(((double) analogInputValue[1] - ((double) _zoomMin
				- (double) _objectiveMin * temp)) / temp);
		int j;

		for (j = 0; j <= (int) ((_objectiveMax - _objectiveMin) / 5); j++) {
			if (zoom >= _objectiveMax - ((j + 1) * 5) && zoom <= _objectiveMax
					- j * 5) {
				i = (int) ((_objectiveMax - _objectiveMin) / 5) - j;
			}

		}

//				printf("Fokus :%d\n", (int) round((double) (array[i] - _focusMin)
//						/ (double) (_focusMax - _focusMin) * (double) 100));
		setfocusto=(int) round((double) (array[i] - _focusMin)
				/ (double) (_focusMax - _focusMin) * (double) 100);
		if (setfocusto<1)setfocusto=1;
		if (setfocusto>100)setfocusto=100;
		setFocusToMm(setfocusto);

		/*setFocusToMm((int) round((double) (array[i] - _focusMin)
				/ (double) (_focusMax - _focusMin) * (double) 100));*/
		fclose(data);
	}
	return setfocusto;
}

void setFocusFromLUTNoWait(int inputzoom) {
	FILE *data;
	data = fopen("LookUpTable.txt", "r");
	ssize_t getlineresult=0;
	//printf("Hello World\n");
	//char intchar[30];
	char* intchar;
	size_t bytetoread=0;
	if (data != NULL) {
		int i;
		int array[(int) ((_objectiveMax - _objectiveMin) / 5) + 1];
		for (i = 0; i <= (int) ((_objectiveMax - _objectiveMin) / 5); i++) {
			//printf("Hello World\n");
			//fread(&array[i], sizeof(int), 1, data);
			getlineresult=getline(&intchar,&bytetoread,data);
			//printf("getlineresult %d --\n",getlineresult);
			array[i]=atoi(intchar);
			//printf("%d --",array[i]);
			//printf("%s --\n",intchar);

		}
		//getAnalogValues();

		int zoom;
		double temp;
		temp = (double) (_zoomMax - _zoomMin) / (double) (_objectiveMax
				- _objectiveMin);
		//zoom = (int) round(((double) analogInputValue[1] - ((double) _zoomMin
		//		- (double) _objectiveMin * temp)) / temp);
		zoom = (int) inputzoom;
		int j;

		for (j = 0; j <= (int) ((_objectiveMax - _objectiveMin) / 5); j++) {
			if (zoom >= _objectiveMax - ((j + 1) * 5) && zoom <= _objectiveMax
					- j * 5) {
				i = (int) ((_objectiveMax - _objectiveMin) / 5) - j;
			}

		}

//				printf("Fokus :%d\n", (int) round((double) (array[i] - _focusMin)
//						/ (double) (_focusMax - _focusMin) * (double) 100));

		setFocusToMmNoWait((int) round((double) (array[i] - _focusMin)
				/ (double) (_focusMax - _focusMin) * (double) 100));
		fclose(data);
	}
}
