/* GStreamer
 * Copyright (C) 2006 David A. Schleef <ds@schleef.org>
 *
 * gstshmsrc.c:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/**
 * SECTION:element-shmsrc
 * @see_also: #GstFileSrc
 *
 * Reads buffers from sequentially named files. If used together with an image
 * decoder, one needs to use the #Gstshmsrc:caps property or a capsfilter
 * to force to caps containing a framerate. Otherwise image decoders send EOS
 * after the first picture.
 *
 * File names are created by replacing "%%d" with the index using printf().
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch shmsrc location="img.%04d.png" index=0 caps="image/png,framerate=\(fraction\)12/1" ! \
 *     pngdec ! ffmpegcolorspace ! theoraenc ! oggmux ! \
 *     filesink location="images.ogg"
 * ]| This pipeline creates a video file "images.ogg" by joining multiple PNG
 * files named img.0000.png, img.0001.png, etc.
 * </refsect2>
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>

//#include <sys/inotify.h>
#include <sys/select.h>

#include <sys/types.h>
#include  <signal.h>
#include <sys/time.h>
#include <unistd.h>

#include "gstshmsrc.h"



static GstFlowReturn gst_shmsrc_create (GstPushSrc * src,
    GstBuffer ** buffer);

static void gst_shmsrc_dispose (GObject * object);

static void gst_shmsrc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_shmsrc_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstCaps *gst_shmsrc_getcaps (GstBaseSrc * src);
static gboolean gst_shmsrc_query (GstBaseSrc * src, GstQuery * query);
static gboolean gst_shmsrc_stop (GstBaseSrc * src);
static gboolean gst_shmsrc_start (GstBaseSrc * src);
static GstStaticPadTemplate gst_shmsrc_pad_template =
GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY);

GST_DEBUG_CATEGORY_STATIC (gst_shmsrc_debug);
#define GST_CAT_DEFAULT gst_shmsrc_debug

static const GstElementDetails gst_shmsrc_details =
GST_ELEMENT_DETAILS ("SHM Source",
    "Source/shm",
    "Subscribe to Shared Memory Regions",
    "Hannes Bistry <bistryATinformatikDOTuniMINUShamburgDOTde>");

enum
{
  ARG_0,
  ARG_CAPS,
  ARG_TOPIC,
  ARG_TRANSPORT
};

#define DEFAULT_LOCATION "%05d"
#define DEFAULT_INDEX 0

GST_BOILERPLATE (Gstshmsrc, gst_shmsrc, GstPushSrc,
    GST_TYPE_PUSH_SRC);

static void
gst_shmsrc_base_init (gpointer g_class)
{

  GstElementClass *gstelement_class = GST_ELEMENT_CLASS (g_class);

  GST_DEBUG_CATEGORY_INIT (gst_shmsrc_debug, "shmsrc", 0,
      "shmsrc element");
  
  gst_element_class_add_pad_template (gstelement_class,
      gst_static_pad_template_get (&gst_shmsrc_pad_template));
  gst_element_class_set_details (gstelement_class, &gst_shmsrc_details);
}

static void
gst_shmsrc_class_init (GstshmsrcClass * klass)
{
  
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstPushSrcClass *gstpushsrc_class = GST_PUSH_SRC_CLASS (klass);
  GstBaseSrcClass *gstbasesrc_class = GST_BASE_SRC_CLASS (klass);

  gobject_class->set_property = gst_shmsrc_set_property;
  gobject_class->get_property = gst_shmsrc_get_property;


  g_object_class_install_property (gobject_class, ARG_CAPS,
      g_param_spec_boxed ("caps", "Caps",
          "Caps describing the format of the data.",
          GST_TYPE_CAPS, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_TOPIC,
      g_param_spec_string ("topic", "Topic",
          "Topic to subscribe to.",
          "Gstshm", G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_TRANSPORT,
      g_param_spec_string ("transport", "Transport",
          "Transport method (raw/compressed/theora/...) .",
          "raw", G_PARAM_READWRITE));



  gobject_class->dispose = gst_shmsrc_dispose;

  gstbasesrc_class->get_caps = gst_shmsrc_getcaps;
  gstbasesrc_class->query = gst_shmsrc_query;
  gstbasesrc_class->start = gst_shmsrc_start;
  gstbasesrc_class->stop = gst_shmsrc_stop;
  gstpushsrc_class->create = gst_shmsrc_create;
  
/*
  if (sizeof (off_t) < 8) {
    GST_LOG ("No large file support, sizeof (off_t) = %" G_GSIZE_FORMAT,
        sizeof (off_t));}
*/
  
}

static void
gst_shmsrc_init (Gstshmsrc * shmsrc,
    GstshmsrcClass * g_class)
{
  shmsrc->successful_read = FALSE;
  shmsrc->topic=g_strdup_printf("Gstshm");
  shmsrc->transport=g_strdup_printf("raw");
  //shmsrc->buffers=NULL;


}

static void
gst_shmsrc_dispose (GObject * object)
{
  Gstshmsrc *src = GST_SHMSRC (object);

  g_free (src->dirname);
  src->dirname = NULL;
  if (src->newcaps)
    gst_caps_unref (src->newcaps);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static GstCaps *
gst_shmsrc_getcaps (GstBaseSrc * src)
{
  Gstshmsrc *multi_file_src = GST_SHMSRC (src);

  GST_DEBUG_OBJECT (src, "returning %" GST_PTR_FORMAT, multi_file_src->newcaps);

  if (multi_file_src->newcaps) {
    return gst_caps_ref (multi_file_src->newcaps);
  } else {
    return gst_caps_new_any ();
  }
}




static gboolean gst_shmsrc_start (GstBaseSrc * src){

  Gstshmsrc *insrc;

  insrc = GST_SHMSRC (src);
  //insrc->end=FALSE;
  insrc->sentbuffers=NULL;
  insrc->data_cond=g_cond_new(); /* Must be initialized somewhere */
  insrc->data_mutex=g_mutex_new();
  insrc->newcaps=NULL;
  insrc->zaehler=30;
  insrc->nenner=1;
  insrc->width=-1;
  insrc->height=-1;
  insrc->index=G_MAXUINT64;
  insrc->caps=NULL; 
  bool connected=FALSE;
  //the shm stuff
  while(!connected){
    try{
      insrc->control= new boost::interprocess::shared_memory_object
      (boost::interprocess::open_only                    //only create
      ,insrc->topic              //name
      ,boost::interprocess::read_write                   //read-write mode
      );
      connected=TRUE;
    }
    catch(boost::interprocess::interprocess_exception &ex){
      std::cout << ex.what() << std::endl;
      sleep(1);
    }
  }
  insrc->region=new boost::interprocess::mapped_region(*insrc->control//What to map
  ,boost::interprocess::read_write //Map it as read-write
  );
  
  void * addr       = insrc->region->get_address();

      //Obtain a pointer to the shared structure
  insrc->tq = static_cast<trace_queue*>(addr);
  return TRUE;
}

static gboolean gst_shmsrc_stop (GstBaseSrc * src){
  Gstshmsrc *insrc;
  //g_print("Stop called\n");
  insrc = GST_SHMSRC (src);
  //insrc->end=TRUE;

  return TRUE;
}

static gboolean
gst_shmsrc_query (GstBaseSrc * src, GstQuery * query)
{
  gboolean res;
  Gstshmsrc *mfsrc;
  //g_print("Shmsrc query now\n");
  mfsrc = GST_SHMSRC (src);

  switch (GST_QUERY_TYPE (query)) {
    case GST_QUERY_POSITION:
    {
      GstFormat format;

      gst_query_parse_position (query, &format, NULL);
      switch (format) {
        case GST_FORMAT_BUFFERS:
        case GST_FORMAT_DEFAULT:
          gst_query_set_position (query, GST_FORMAT_BUFFERS, mfsrc->index);
          res = TRUE;
          break;
        default:
          res = GST_BASE_SRC_CLASS (parent_class)->query (src, query);
          break;
      }
      break;
    }
    default:
      res = GST_BASE_SRC_CLASS (parent_class)->query (src, query);
      break;
  }
  return res;
}





static void
gst_shmsrc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstshmsrc *src = GST_SHMSRC (object);

  switch (prop_id) {
    case ARG_CAPS:
    {
      const GstCaps *caps = gst_value_get_caps (value);
      GstCaps *new_caps;

      if (caps == NULL) {
        new_caps = gst_caps_new_any ();
      } else {
        new_caps = gst_caps_copy (caps);
      }
      gst_caps_replace (&src->caps, new_caps);
      gst_pad_set_caps (GST_BASE_SRC_PAD (src), new_caps);
    }
      break;
    case ARG_TOPIC:
    {
      if (src->topic)g_free(src->topic);
      src->topic=g_strdup(g_value_get_string (value)); 
    }
      break;
    case ARG_TRANSPORT:
    {
      if (src->transport)g_free(src->transport);
      src->transport=g_strdup(g_value_get_string (value)); 
    }
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_shmsrc_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstshmsrc *src = GST_SHMSRC (object);

  switch (prop_id) {
    case ARG_CAPS:
      gst_value_set_caps (value, src->caps);
      break;
    case ARG_TOPIC:
      g_value_set_string (value, src->topic);
      break;
    case ARG_TRANSPORT:
      g_value_set_string (value, src->transport);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


static GstFlowReturn
gst_shmsrc_create (GstPushSrc * src, GstBuffer ** buffer)
{
  Gstshmsrc *insrc;
  gsize size;
  gchar *data;
  gboolean ret;
  GError *error = NULL;

  insrc = GST_SHMSRC (src);
  int result;
  //if (insrc->end){
  //      g_print("Will now return Unexpected\n");
//	return GST_FLOW_UNEXPECTED; 
//  }
  //g_print("Will lock mutex now\n");
  g_mutex_lock (insrc->data_mutex);
  //kicking old buffers
  /*GList* temp=insrc->sentbuffers;
  while(temp){
    if (((GstMiniObject*)(temp->data))->refcount==1){
	//search for Data pointer
	//for (std::vector<sensor_msgs::ImageConstPtr>::iterator it = insrc->msglist.begin() ; it != insrc->msglist.end(); ++it){
	  //g_print("Considering to delete %p\n",(&((*it)->data[0])));
	//  if ((&((*it)->data[0]))==((GstBuffer*)(temp->data))->data) {
	//	insrc->msglist.erase(it);
	//	break;
	//  }
	//}

	gst_buffer_unref((GstBuffer*)temp->data);
	insrc->sentbuffers=g_list_delete_link(insrc->sentbuffers,temp);
	temp=insrc->sentbuffers;
    }	
    else temp=g_list_next(temp);
  }*/
  GstBuffer* newbuf=NULL;
  
   std::vector< boost::interprocess::mapped_region* >::iterator it3 = insrc->regionlist.begin(); 
  std::vector<boost::interprocess::shared_memory_object*>::iterator it2 = insrc->msglist.begin();
  for (std::vector<GstBuffer*>::iterator it = insrc->buflist.begin(); it != insrc->buflist.end(); )    {
    //g_print("Loop checking %p whether we still need it\n",(*it));
    if (((GstBuffer*)(*it))->mini_object.refcount==1){
      //g_print("GstBuffer %p seems not to be used anymore, dropping\n",(*it));
      if(!newbuf)newbuf=((GstBuffer*)(*it));
      else gst_buffer_unref((GstBuffer*)(*it));
      delete (*it3);     
      delete (*it2);
      it=insrc->buflist.erase(it);
      it2=insrc->msglist.erase(it2);
      it3=insrc->regionlist.erase(it3);
    }
    else{ 
      ++it;
      if (it2 != insrc->msglist.end())++it2;
      if (it3 != insrc->regionlist.end())++it3;     
    }
  }  
  
  if (newbuf==NULL)newbuf=gst_buffer_new();
  boost::interprocess::shared_memory_object* shm;
  boost::interprocess::mapped_region* regionp;
  {
    //g_print ("Shmsrc will now lock scoped mutex\n"); 
    boost::interprocess::scoped_lock<boost::interprocess::interprocess_mutex> lock(insrc->tq->mutex);
    //sleep(5);
    if (insrc->tq->currentframe==insrc->index) {
      //g_print ("Shmsrc will now definitely lock semaphore\n");      
      insrc->tq->cond_empty.wait(lock);//wait until data
    }
    insrc->index=insrc->tq->currentframe;
    //g_print ("Gstsrc will now take buffer of %llu.\n",(long long unsigned int)insrc->index);
    sprintf(insrc->shmname,"%s%llu",insrc->topic,(long long unsigned int)insrc->index);
    shm=new boost::interprocess::shared_memory_object
      (boost::interprocess::open_only                    //only create
      ,insrc->shmname             //name
      ,boost::interprocess::read_only                   //read-write mode
      );

    regionp=new boost::interprocess::mapped_region
         (*shm                       //What to map
         ,boost::interprocess::read_only //Map it as read-write
         );    
    if ((*((int*)(regionp->get_address()))==1)||(!insrc->caps)){//get caps from buf
      if (insrc->caps)gst_caps_unref(insrc->caps);
      //insrc->caps=gst_caps_from_string(regionp->get_address()+1000+virtbuf->size);//todo memcheck
      insrc->caps=gst_caps_from_string(insrc->tq->caps);//todo memcheck
    }
  }

  GstBuffer* virtbuf=(GstBuffer*)(regionp->get_address()+4);
  GST_BUFFER_DATA(newbuf)=(regionp->get_address()+1000);

  gst_buffer_copy_metadata(newbuf,virtbuf,(GST_BUFFER_COPY_FLAGS|
      GST_BUFFER_COPY_TIMESTAMPS));
  newbuf->size=virtbuf->size;
  gst_buffer_ref(newbuf); 
  

  gst_buffer_set_caps(newbuf,insrc->caps);
  
  //test buffer
  //int sum;
  //g_print("Checking data regio  %p, and base of shmthing %p\n",newbuf->data,regionp->get_address()); 
  //for (int i=0; i<newbuf->size;i++)sum+=newbuf->data[i];
  //g_print("Sum is %d\n",sum);
  insrc->msglist.push_back(shm);
  insrc->buflist.push_back(newbuf); 
  insrc->regionlist.push_back(regionp); 
  *buffer=newbuf;
  //while (!insrc->buffer)
    //g_cond_wait (insrc->data_cond, insrc->data_mutex);
  //data = current_data;
  //current_data = NULL;
  //g_print("Awaked from waiting for buffer\n");

  /*
  if (!insrc->current){
        g_print("Awaked, but no buffer present\n");
        g_mutex_unlock (insrc->data_mutex);
	return GST_FLOW_UNEXPECTED; 
  }
  */
  insrc->successful_read = TRUE;
  //GST_DEBUG_OBJECT (insrc, "read file \"%s\".", filename);
  //*buffer = gst_buffer_ref(insrc->current);
  //insrc->sentbuffers=g_list_append(insrc->sentbuffers,insrc->current);
  //insrc->current=NULL;
  g_mutex_unlock (insrc->data_mutex);
  return GST_FLOW_OK;
}




//static gboolean init_shmsrc (GstPlugin * plugin){
//return gst_element_register (plugin, "shmsrc", GST_RANK_NONE, GST_TYPE_SHMSRC);}
//GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,GST_VERSION_MINOR,"shmsrc","Template shmsrc",init_shmsrc,VERSION,"LGPL","GStreamer","http:gstreamer.net")
