/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2009 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-videosplitter
 *
 * Split stream of video in 4 Substreams.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m videotestsrc ! videosplitter name=spl ! videojoin name=join ! ffmpegcolorspace ! ximagesink spl. ! join
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include "gstvideosplitter.h"
#include <string.h>

GST_DEBUG_CATEGORY_STATIC (gst_videosplitter_debug);
#define GST_CAT_DEFAULT gst_videosplitter_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_OVERLAP,
  PROP_OPMODE,
  PROP_NUMOUTPUTS,
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src%d",
    GST_PAD_SRC,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS ("ANY")
    );

GST_BOILERPLATE (Gstvideosplitter, gst_videosplitter, GstElement,
    GST_TYPE_ELEMENT);

static void gst_videosplitter_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_videosplitter_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstCaps * gst_videosplittersrc_get_caps (GstPad *pad);
static gboolean gst_videosplitter_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_videosplitter_chain (GstPad * pad, GstBuffer * buf);

static GstPad *
gst_videosplitter_request_new_pad (GstElement * element, GstPadTemplate * temp,
    const gchar * name);
static void
gst_videosplitter_release_pad (GstElement * element, GstPad * pad);



#define VIDEOSPLITER_OP_MODE (videosplitter_type_op_mode ())
static GType
videosplitter_type_op_mode (void)
{
  static GType op_mode = 0;
  static const GEnumValue functions[] = {
    {HORIZONTAL, "Split image horizontally (efficient, ZERO-copy)", "horizontal"},
    {VERTICAL, "Split image vertically (needs copy)", "vertival"},
    {0, NULL, NULL}
  };

  if (!op_mode) {
    op_mode =
        g_enum_register_static ("VideosplitterOpMode", functions);
  }
  return op_mode;
}


/* GObject vmethod implementations */

static void
gst_videosplitter_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "videosplitter",
    "Gst-Plugins-Robotix",
    "Splits stream of video into Substreams",
    "Hannes Bistry <<hannesb at gmx.de>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the videosplitter's class */
static void
gst_videosplitter_class_init (GstvideosplitterClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_videosplitter_set_property;
  gobject_class->get_property = gst_videosplitter_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,PROP_OVERLAP,
      g_param_spec_int ("overlap", "Overlap",
          "How big is the overlapping area at least", G_MININT,G_MAXINT,
          0, (GParamFlags)G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, PROP_OPMODE,
      g_param_spec_enum ("opmode", "Opmode",
          "Wether to split vertically or horizontally", VIDEOSPLITER_OP_MODE,
          HORIZONTAL, (GParamFlags)G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,PROP_NUMOUTPUTS,
      g_param_spec_uint ("numoutputs", "Numoutputs",
          "Number of output pads currently active", 0,G_MAXUINT,
          0, (GParamFlags)G_PARAM_READABLE));
  gstelement_class->request_new_pad =
      GST_DEBUG_FUNCPTR (gst_videosplitter_request_new_pad);
  gstelement_class->release_pad =
      GST_DEBUG_FUNCPTR (gst_videosplitter_release_pad);
}

void
gst_videosplitter_update_overlap(Gstvideosplitter * filter,int overlap){

  g_print ("Calculating Overlap min %d and hight %d width %d unith %d unitw %d\n",filter->overlap,filter->height,filter->width,filter->unith,filter->unitw);     
  int temp=overlap;
  filter->overlap=overlap;
  if ((filter->height>0)&&(filter->num_srcpads>1)){
    //if equally dividable && result is a multiple of unit && overlap is a multiple of unit && search has not finished
    while (((((filter->height+((filter->num_srcpads-1)*temp))%filter->num_srcpads)!=0)||((((filter->height+((filter->num_srcpads-1)*temp))/filter->num_srcpads)%filter->unith)!=0)||((temp%filter->unith)!=0))&&(temp<filter->height))temp++;
    filter->overlap_h = temp;
    filter->heightsmall=(filter->height+((filter->num_srcpads-1)*filter->overlap_h))/filter->num_srcpads;
    g_print ("Found solution for horizontal splitting with Overlap %d and hight %d\n",filter->overlap_h,filter->heightsmall);   
    if ((filter->heightsmall>filter->height)||(filter->overlap_h>filter->height)){
      filter->overlap_h=filter->heightsmall=filter->height;
      g_print ("Warning, solution exceeds bounds of image\n");  
    }
  }

  if ((filter->width>0)&&(filter->num_srcpads>1)){
    while (((((filter->width+((filter->num_srcpads-1)*temp))%filter->num_srcpads)!=0)||((((filter->width+((filter->num_srcpads-1)*temp))/filter->num_srcpads)%filter->unitw)!=0)||((temp%filter->unitw)!=0))&&(temp<2*filter->width)) temp++;
    filter->overlap_v = temp;
    filter->widthsmall=(filter->width+((filter->num_srcpads-1)*filter->overlap_v))/filter->num_srcpads;
    g_print ("Found solution for vertical splitting with Overlap %d and width %d\n",filter->overlap_v,filter->widthsmall);   
    if ((filter->widthsmall>filter->width)||(filter->overlap_v>filter->width)){
      filter->overlap_v=filter->widthsmall=filter->width;
      g_print ("Warning, solution exceeds bounds of image\n");  
    }
  }
  if (filter->num_srcpads==1){
    filter->overlap_h=0;
    filter->overlap_v=0;
    filter->heightsmall=filter->height;
    filter->widthsmall=filter->width;
  }
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_videosplitter_init (Gstvideosplitter * filter,
    GstvideosplitterClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_videosplitter_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_videosplitter_chain));
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);

  filter->silent = FALSE;
  filter->height = 0;
  filter->drop = 0;
}

gboolean gst_videosplitter_modifycaps(GObject * object)
{
  Gstvideosplitter *filter = GST_VIDEOSPLITTER (object);
  if (filter->srccaps) gst_caps_unref (filter->srccaps);
  filter->srccaps=NULL;
  if (GST_IS_CAPS(filter->sinkcaps))filter->srccaps=gst_caps_copy(filter->sinkcaps);
  else return FALSE;
  g_print("Videosplitter: Before Src caps are %s.\n",gst_caps_to_string(filter->srccaps));
  GstStructure *structure=gst_caps_get_structure(filter->srccaps,0);
  if (!structure) g_print("Scheisse!\n");
  //gst_structure_fixate_field_nearest_int (structure,"width",filter->widthsmall);
  if (filter->opmode==HORIZONTAL){
    gst_structure_set(structure,"height",G_TYPE_INT,filter->heightsmall,NULL);
    gst_structure_set(structure,"overlap_h",G_TYPE_INT,filter->overlap_h,NULL);
    gst_structure_set(structure,"overlap_v",G_TYPE_INT,0,NULL);
  }
  else if(filter->opmode==VERTICAL){
    gst_structure_set(structure,"width",G_TYPE_INT,filter->widthsmall,NULL);
    gst_structure_set(structure,"overlap_v",G_TYPE_INT,filter->overlap_v,NULL);
    gst_structure_set(structure,"overlap_h",G_TYPE_INT,0,NULL);
  }
  g_print("Videosplitter: Now Source caps are %s.\n",gst_caps_to_string(filter->srccaps));
  return TRUE;
}

static void
gst_videosplitter_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstvideosplitter *filter = GST_VIDEOSPLITTER (object);
  gint temp;
  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_OVERLAP:
      if (GST_STATE(filter)==GST_STATE_NULL){
        temp=g_value_get_int (value);
        gst_videosplitter_update_overlap(filter,temp);
        gst_videosplitter_modifycaps(object);//to take effect
      }
      break;
    case PROP_OPMODE:
      if (GST_STATE(filter)==GST_STATE_NULL){
      	filter->opmode=(VideosplitterOpMode)g_value_get_enum(value);
        gst_videosplitter_modifycaps(object);
	}
      else printf("WARNING!!! Changing op-mode only allowed when stopped\n");
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_videosplitter_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstvideosplitter *filter = GST_VIDEOSPLITTER (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_OVERLAP:
      g_value_set_int (value, filter->overlap);
      break;
    case PROP_OPMODE:
      g_value_set_enum(value, filter->opmode);
      break;
    case PROP_NUMOUTPUTS:
      g_value_set_uint (value, filter->num_srcpads);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */


static GstCaps *
gst_videosplittersrc_get_caps (GstPad *pad){
  //g_print("PIPgen: Start get caps!!\n");
  Gstvideosplitter *filter;
  filter = GST_VIDEOSPLITTER (gst_pad_get_parent (pad));

  if (filter&&filter->srccaps)return gst_caps_copy(filter->srccaps);
  else return GST_CAPS_ANY;
  

}
/* this function handles the link with other elements */
static gboolean
gst_videosplitter_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstvideosplitter *filter;
  filter = GST_VIDEOSPLITTER (gst_pad_get_parent (pad));
  if (pad!=filter->sinkpad) return TRUE;


  GstStructure *structure = gst_caps_get_structure (caps, 0);
  gboolean bppset=FALSE;
  const gchar *mime;
  mime = gst_structure_get_name (structure);    
  if (gst_structure_get_int(structure,"bpp",&(filter->bpp))) bppset=TRUE;

if (0 == strcmp (mime, "video/x-raw-yuv")){ 
	guint32 fourccval;
	if (!gst_structure_has_field(structure,"format")){
		g_print ("YUV No format field");
		return FALSE;
	}  
	if (!gst_structure_get_fourcc(structure,"format",&fourccval)){
		g_print ("FourCC could not be read");
		return FALSE;
	}
	if (fourccval==GST_MAKE_FOURCC ('Y', '8', '0', '0')){
	  if (!bppset) filter->bpp=8;
	  bppset=TRUE;
	  filter->unitw=filter->unith=1;
	}
	else if (fourccval==GST_MAKE_FOURCC ('Y', '4', '2', '2')){
	  if (!bppset) filter->bpp=16;
	  bppset=TRUE;
	  filter->unitw=2;
	  filter->unith=1;
	}
	else if (fourccval==GST_MAKE_FOURCC ('Y', 'U', 'Y', '2')){
	  if (!bppset) filter->bpp=16;
	  bppset=TRUE;
	  filter->unitw=2;
	  filter->unith=1;
	}
}
else if (0 == strcmp (mime, "video/x-raw-gray")){
filter->unitw=filter->unith=1;
}
else if (0 == strcmp (mime, "video/x-raw-rgb")){
filter->unitw=filter->unith=1;
}

  if (!gst_structure_get_int(structure,"width",&(filter->width))){
	g_print ("WARNING: Width not set!");
	return FALSE;
  }
  if (!gst_structure_get_int(structure,"height",&(filter->height))){
	g_print ("WARNING Height not set!");
	return FALSE;
  }

//calculate height of output
gst_videosplitter_update_overlap(filter,filter->overlap);
//copy caps and modifie them
  filter->sinkcaps=caps;
  gst_videosplitter_modifycaps((GObject*)filter);
//
  filter->bytesperline=(filter->width*filter->bpp)/8;
  g_print ("Bytesperline %d ",filter->bytesperline);
  gboolean result=TRUE;
  GList *tmp;
  for (tmp = filter->srcpads; tmp; tmp = g_list_next (tmp)){
    result=result&&(gst_pad_set_caps (GST_PAD(tmp->data), filter->srccaps));
  }
  return result;
}



static GstPad *
gst_videosplitter_request_new_pad (GstElement * element, GstPadTemplate * temp,
    const gchar * name)
{
  gchar* tmp;
  GstPad* tempPad;
  Gstvideosplitter *filter;
  filter = GST_VIDEOSPLITTER (element);

  //if (!name) tmp = g_strdup_printf ("src%d", filter->num_srcpads);
  //else tmp=g_strdup_printf ("%s", name);
  tmp = g_strdup_printf ("src%d", filter->num_srcpads);
  tempPad = gst_pad_new_from_static_template (&src_factory, tmp);
  g_free (tmp);

  filter->srcpads=g_list_append(filter->srcpads,tempPad);
  filter->num_srcpads++;
  gst_videosplitter_update_overlap(filter,filter->overlap);
  gst_videosplitter_modifycaps((GObject*)filter);
  gst_pad_set_getcaps_function (tempPad,
      GST_DEBUG_FUNCPTR (gst_videosplittersrc_get_caps));
  gst_element_add_pad (GST_ELEMENT (filter), tempPad);
  return tempPad;
}



static void
gst_videosplitter_release_pad (GstElement * element, GstPad * pad)
{

g_print("releasing pads not supported yet:)\n");
for(;;);
}







/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_videosplitter_chain (GstPad * pad, GstBuffer * buf)
{
  Gstvideosplitter *filter;
  filter = GST_VIDEOSPLITTER (GST_OBJECT_PARENT (pad));
  GstFlowReturn result=GST_FLOW_OK;
//  filter->drop++;
//  if (filter->drop%100!=0){
//	gst_buffer_unref(buf);
//	return result;
//  }
  int counter=0,h;
  GList *tmp;
  GstBuffer *tmpbuf;
  if (filter->opmode==HORIZONTAL){
    for (tmp=filter->srcpads;((tmp)&&(result==GST_FLOW_OK));tmp=g_list_next(tmp)){
      tmpbuf=gst_buffer_create_sub(buf,counter*(filter->bytesperline*(filter->heightsmall-filter->overlap_h)),filter->bytesperline*filter->heightsmall);
    gst_buffer_set_caps(tmpbuf,filter->srccaps);
    tmpbuf->offset=buf->offset;
    tmpbuf->duration=buf->duration;
    tmpbuf->timestamp=buf->timestamp;
    counter++;
    result=gst_pad_push (GST_PAD(tmp->data), tmpbuf);
    }
    gst_buffer_unref(buf);   
    return result;
 }
  if (filter->opmode==VERTICAL){
    //gint nr=0;
    for (tmp=filter->srcpads;(tmp&&(result==GST_FLOW_OK));tmp=g_list_next(tmp)){
      if (GST_FLOW_OK==(result=gst_pad_alloc_buffer(GST_PAD(tmp->data),buf->offset,(filter->widthsmall*filter->height*(filter->bpp/8)),filter->srccaps,&tmpbuf))){
        for (h=0;h<filter->height;h++){
          //g_print("trying to copy counter %d and h %d\n",counter,h);
	  memcpy(tmpbuf->data+(h*filter->widthsmall*(filter->bpp/8)),buf->data+(h*filter->bytesperline)+(counter*(filter->bpp/8)*(filter->widthsmall-filter->overlap_v)),filter->widthsmall*(filter->bpp/8));
        }
	tmpbuf->timestamp=buf->timestamp;
	tmpbuf->duration=buf->duration;
        result=gst_pad_push (GST_PAD(tmp->data), tmpbuf);
      }
      counter++;
    }
    gst_buffer_unref(buf);   
    return result;
 }
 return GST_FLOW_ERROR;
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
videosplitter_init (GstPlugin * videosplitter)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template videosplitter' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_videosplitter_debug, "videosplitter",
      0, "Splits stream of video into 4 Substreams");

  return gst_element_register (videosplitter, "videosplitter", GST_RANK_NONE,
      GST_TYPE_VIDEOSPLITTER);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstvideosplitter"
#endif

/* gstreamer looks for this structure to register videosplitters
 *
 * exchange the string 'Template videosplitter' with your videosplitter description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "videosplitter",
    "Splits stream of video into 4 Substreams",
    videosplitter_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
